VERSION = (0, 9, 0, "final")

if VERSION[-1] != "final":  # pragma: no cover
    __version__ = ".".join(map(str, VERSION))
else:  # pragma: no cover
    __version__ = ".".join(map(str, VERSION[:-1]))



class FieldRegistry(object):
    _registry = {}

    def __init__(self, fieldcls):
        self._fieldcls = fieldcls

    def add_field(self, model, field):
        reg = self.__class__._registry.setdefault(self._fieldcls, {}).setdefault(
            model, []
        )
        reg.append(field)

    def get_fields(self, model):
        return self.__class__._registry.setdefault(self._fieldcls, {}).get(model, [])

    def __contains__(self, model):
        return model in self.__class__._registry.setdefault(self._fieldcls, {})



from functools import partial
from django.db.models import signals

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object

from logs import registration, settings
from logs.models import fields
from logs.models.managers import AuditLogManager


def _disable_audit_log_managers(instance):
    for attr in dir(instance):
        try:
            if isinstance(getattr(instance, attr), AuditLogManager):
                getattr(instance, attr).disable_tracking()
        except AttributeError:
            pass


def _enable_audit_log_managers(instance):
    for attr in dir(instance):
        try:
            if isinstance(getattr(instance, attr), AuditLogManager):
                getattr(instance, attr).enable_tracking()
        except AttributeError:
            pass


class UserLoggingMiddleware(MiddlewareMixin):
    def process_request(self, request):
        if settings.DISABLE_AUDIT_LOG:
            return
        if not request.method in ("GET", "HEAD", "OPTIONS", "TRACE"):
            if hasattr(request, "user") and request.user.is_authenticated:
                user = request.user
            else:
                user = None
            session = request.session.session_key
            update_pre_save_info = partial(self._update_pre_save_info, user, session)
            update_post_save_info = partial(self._update_post_save_info, user, session)
            signals.pre_save.connect(
                update_pre_save_info,
                dispatch_uid=(
                    self.__class__,
                    request,
                ),
                weak=False,
            )
            signals.post_save.connect(
                update_post_save_info,
                dispatch_uid=(
                    self.__class__,
                    request,
                ),
                weak=False,
            )

    def process_response(self, request, response):
        if settings.DISABLE_AUDIT_LOG:
            return
        signals.pre_save.disconnect(
            dispatch_uid=(
                self.__class__,
                request,
            )
        )
        signals.post_save.disconnect(
            dispatch_uid=(
                self.__class__,
                request,
            )
        )
        return response

    def process_exception(self, request, exception):
        if settings.DISABLE_AUDIT_LOG:
            return None
        signals.pre_save.disconnect(
            dispatch_uid=(
                self.__class__,
                request,
            )
        )
        signals.post_save.disconnect(
            dispatch_uid=(
                self.__class__,
                request,
            )
        )
        return None

    def _update_pre_save_info(self, user, session, sender, instance, **kwargs):
        registry = registration.FieldRegistry(fields.LastUserField)
        if sender in registry:
            for field in registry.get_fields(sender):
                setattr(instance, field.name, user)

        registry = registration.FieldRegistry(fields.LastSessionKeyField)
        if sender in registry:
            for field in registry.get_fields(sender):
                setattr(instance, field.name, session)

    def _update_post_save_info(
        self, user, session, sender, instance, created, **kwargs
    ):
        if created:
            registry = registration.FieldRegistry(fields.CreatingUserField)
            if sender in registry:
                for field in registry.get_fields(sender):
                    setattr(instance, field.name, user)
                    _disable_audit_log_managers(instance)
                    instance.save()
                    _enable_audit_log_managers(instance)

            registry = registration.FieldRegistry(fields.CreatingSessionKeyField)
            if sender in registry:
                for field in registry.get_fields(sender):
                    setattr(instance, field.name, session)
                    _disable_audit_log_managers(instance)
                    instance.save()
                    _enable_audit_log_managers(instance)


class JWTAuthMiddleware(MiddlewareMixin):
    """
    Convenience middleware for users of django-rest-framework-jwt.
    Fixes issue https://github.com/GetBlimp/django-rest-framework-jwt/issues/45
    """

    def get_user_jwt(self, request):
        from rest_framework.request import Request
        from rest_framework.exceptions import AuthenticationFailed
        from django.utils.functional import SimpleLazyObject
        from django.contrib.auth.middleware import get_user
        from rest_framework_jwt.authentication import JSONWebTokenAuthentication

        user = get_user(request)
        if user.is_authenticated:
            return user
        try:
            user_jwt = JSONWebTokenAuthentication().authenticate(Request(request))
            if user_jwt is not None:
                return user_jwt[0]
        except AuthenticationFailed:
            pass
        return user

    def process_request(self, request):
        from django.utils.functional import SimpleLazyObject

        assert hasattr(
            request, "session"
        ), """The Django authentication middleware requires session middleware to be installed.
         Edit your MIDDLEWARE setting to insert 'django.contrib.sessions.middleware.SessionMiddleware'."""

        request.user = SimpleLazyObject(lambda: self.get_user_jwt(request))



import copy
import datetime
from functools import partial
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.conf import settings

from logs.models.fields import LastUserField
from logs import settings as local_settings


try:
    from django.utils.timezone import now as datetime_now

    assert datetime_now
except ImportError:
    import datetime

    datetime_now = datetime.datetime.now


class LogEntryObjectDescriptor(object):
    def __init__(self, model):
        self.model = model

    def __get__(self, instance, owner):
        kwargs = dict(
            (f.attname, getattr(instance, f.attname))
            for f in self.model._meta.fields
            if hasattr(instance, f.attname)
        )
        return self.model(**kwargs)


class AuditLogManager(models.Manager):
    def __init__(
        self,
        model,
        attname,
        instance=None,
    ):
        super(AuditLogManager, self).__init__()
        self.model = model
        self.instance = instance
        self.attname = attname
        # set a hidden attribute on the  instance to control wether we should track changes
        if instance is not None and not hasattr(instance, "__is_%s_enabled" % attname):
            setattr(instance, "__is_%s_enabled" % attname, True)

    def enable_tracking(self):
        if self.instance is None:
            raise ValueError(
                "Tracking can only be enabled or disabled "
                "per model instance, not on a model class"
            )
        setattr(self.instance, "__is_%s_enabled" % self.attname, True)

    def disable_tracking(self):
        if self.instance is None:
            raise ValueError(
                "Tracking can only be enabled or disabled "
                "per model instance, not on a model class"
            )
        setattr(self.instance, "__is_%s_enabled" % self.attname, False)

    def is_tracking_enabled(self):
        if local_settings.DISABLE_AUDIT_LOG:
            return False
        if self.instance is None:
            raise ValueError(
                "Tracking can only be enabled or disabled "
                "per model instance, not on a model class"
            )
        return getattr(self.instance, "__is_%s_enabled" % self.attname)

    def get_queryset(self):
        if self.instance is None:
            return super(AuditLogManager, self).get_queryset()

        f = {self.instance._meta.pk.name: self.instance.pk}
        return super(AuditLogManager, self).get_queryset().filter(**f)


class AuditLogDescriptor(object):
    def __init__(self, model, manager_class, attname):
        self.model = model
        self.manager_class = manager_class
        self.attname = attname

    def __get__(self, instance, owner):
        if instance is None:
            return self.manager_class(self.model, self.attname)
        return self.manager_class(self.model, self.attname, instance)


class AuditLog(object):
    manager_class = AuditLogManager

    def __init__(self, exclude=[]):
        self._exclude = exclude

    def contribute_to_class(self, cls, name):
        self.manager_name = name
        models.signals.class_prepared.connect(self.finalize, sender=cls)

    def create_log_entry(self, instance, action_type):
        manager = getattr(instance, self.manager_name)
        attrs = {}
        for field in instance._meta.fields:
            if field.attname not in self._exclude:
                attrs[field.attname] = getattr(instance, field.attname)
        manager.create(action_type=action_type, **attrs)

    def post_save(self, instance, created, **kwargs):
        # ignore if it is disabled
        if getattr(instance, self.manager_name).is_tracking_enabled():
            self.create_log_entry(instance, created and "I" or "U")

    def post_delete(self, instance, **kwargs):
        # ignore if it is disabled
        if getattr(instance, self.manager_name).is_tracking_enabled():
            self.create_log_entry(instance, "D")

    def finalize(self, sender, **kwargs):
        log_entry_model = self.create_log_entry_model(sender)

        models.signals.post_save.connect(self.post_save, sender=sender, weak=False)
        models.signals.post_delete.connect(self.post_delete, sender=sender, weak=False)

        descriptor = AuditLogDescriptor(
            log_entry_model, self.manager_class, self.manager_name
        )
        setattr(sender, self.manager_name, descriptor)

    def copy_fields(self, model):
        """
        Creates copies of the fields we are keeping
        track of for the provided model, returning a
        dictionary mapping field name to a copied field object.
        """
        fields = {"__module__": model.__module__}

        for field in model._meta.fields:
            if not field.name in self._exclude:
                field = copy.deepcopy(field)

                if isinstance(field, models.AutoField):
                    # we replace the AutoField of the original model
                    # with an IntegerField because a model can
                    # have only one autofield.

                    field.__class__ = models.IntegerField

                if field.primary_key:
                    field.serialize = True

                # OneToOne fields should really be tracked
                # as ForeignKey fields
                if isinstance(field, models.OneToOneField):
                    field.__class__ = models.ForeignKey

                if field.primary_key or field.unique:
                    # unique fields of the original model
                    # can not be guaranteed to be unique
                    # in the audit log entry but they
                    # should still be indexed for faster lookups.

                    field.primary_key = False
                    field._unique = False
                    field.db_index = True

                if field.remote_field and field.remote_field.related_name:
                    field.remote_field.related_name = "_auditlog_{}_{}".format(
                        model._meta.model_name, field.remote_field.related_name
                    )
                elif field.remote_field:
                    try:
                        if field.remote_field.get_accessor_name():
                            field.remote_field.related_name = "_auditlog_{}_{}".format(
                                model._meta.model_name,
                                field.remote_field.get_accessor_name(),
                            )
                    except e:
                        pass

                fields[field.name] = field

        return fields

    def get_logging_fields(self, model):
        """
        Returns a dictionary mapping of the fields that are used for
        keeping the acutal audit log entries.
        """
        rel_name = "_%s_audit_log_entry" % model._meta.object_name.lower()

        def entry_instance_to_unicode(log_entry):
            try:
                result = "%s: %s %s at %s" % (
                    model._meta.object_name,
                    log_entry.object_state,
                    log_entry.get_action_type_display().lower(),
                    log_entry.action_date,
                )
            except AttributeError:
                result = "%s %s at %s" % (
                    model._meta.object_name,
                    log_entry.get_action_type_display().lower(),
                    log_entry.action_date,
                )
            return result

        action_user_field = LastUserField(related_name=rel_name, editable=False)

        # check if the manager has been attached to auth user model
        if [model._meta.app_label, model.__name__] == getattr(
            settings, "AUTH_USER_MODEL", "auth.User"
        ).split("."):
            action_user_field = LastUserField(
                related_name=rel_name, editable=False, to="self"
            )

        return {
            "action_id": models.AutoField(primary_key=True),
            "action_date": models.DateTimeField(
                default=datetime_now, editable=False, blank=False
            ),
            "action_user": action_user_field,
            "action_type": models.CharField(
                max_length=1,
                editable=False,
                choices=(
                    ("I", _("Created")),
                    ("U", _("Changed")),
                    ("D", _("Deleted")),
                ),
            ),
            "object_state": LogEntryObjectDescriptor(model),
            "__unicode__": entry_instance_to_unicode,
        }

    def get_meta_options(self, model):
        """
        Returns a dictionary of Meta options for the
        autdit log model.
        """
        result = {
            "ordering": ("-action_date",),
            "app_label": model._meta.app_label,
        }
        from django.db.models.options import DEFAULT_NAMES

        if "default_permissions" in DEFAULT_NAMES:
            result.update({"default_permissions": ()})
        return result

    def create_log_entry_model(self, model):
        """
        Creates a log entry model that will be associated with
        the model provided.
        """

        attrs = self.copy_fields(model)
        attrs.update(self.get_logging_fields(model))
        attrs.update(Meta=type(str("Meta"), (), self.get_meta_options(model)))
        name = str("%sAuditLogEntry" % model._meta.object_name)
        return type(name, (models.Model,), attrs)



import os
import sys
from setuptools import setup, find_packages

import logs
from logs import VERSION, __version__

if VERSION[-1] == "final":
    STATUS = ["Development Status :: 5 - Production/Stable"]
elif "beta" in VERSION[-1]:
    STATUS = ["Development Status :: 4 - Beta"]
else:
    STATUS = ["Development Status :: 3 - Alpha"]


def get_readme():
    try:
        return open(os.path.join(os.path.dirname(__file__), "README.rst")).read()
    except IOError:
        return ""


setup(
    name="django-logs",
    version=__version__,
    packages=find_packages(exclude=["testproject"]),
    description="Logs for django models",
    long_description=get_readme(),
    url="https://github.com/null-none/django-logs",
    download_url="https://github.com/null-none/django-logs",
    include_package_data=True,
    zip_safe=False,
    classifiers=STATUS
    + [
        "Environment :: Plugins",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
    ],
)



from django.db import models
from django.db.models import Model
from logs.models.fields import (
    CreatingUserField,
    CreatingSessionKeyField,
    LastUserField,
    LastSessionKeyField,
)


class AuthStampedModel(Model):
    """
    An abstract base class model that provides auth and session information
    fields.
    """

    created_by = CreatingUserField(
        verbose_name="created by",
        related_name="created_%(app_label)s_%(class)s_set",
        on_delete=models.SET_NULL,
    )
    created_with_session_key = CreatingSessionKeyField(_("created with session key"))
    modified_by = LastUserField(
        verbose_name="modified by",
        related_name="modified_%(app_label)s_%(class)s_set",
        on_delete=models.SET_NULL,
    )
    modified_with_session_key = LastSessionKeyField("modified with session key")

    class Meta:
        abstract = True



from django.db import models
from django.conf import settings

from logs import registration


class LastUserField(models.ForeignKey):
    """
    A field that keeps the last user that saved an instance
    of a model. None will be the value for AnonymousUser.
    """

    def __init__(
        self,
        to=getattr(settings, "AUTH_USER_MODEL", "auth.User"),
        on_delete=models.SET_NULL,
        null=True,
        editable=False,
        **kwargs
    ):
        super(LastUserField, self).__init__(
            to=to, on_delete=on_delete, null=null, editable=editable, **kwargs
        )

    def contribute_to_class(self, cls, name):
        super(LastUserField, self).contribute_to_class(cls, name)
        registry = registration.FieldRegistry(self.__class__)
        registry.add_field(cls, self)


class LastSessionKeyField(models.CharField):
    """
    A field that keeps a reference to the last session key that was used to access the model.
    """

    def __init__(self, max_length=40, null=True, editable=False, **kwargs):
        super(LastSessionKeyField, self).__init__(
            max_length=40, null=null, editable=editable, **kwargs
        )

    def contribute_to_class(self, cls, name):
        super(LastSessionKeyField, self).contribute_to_class(cls, name)
        registry = registration.FieldRegistry(self.__class__)
        registry.add_field(cls, self)


class CreatingUserField(LastUserField):
    """
    A field that keeps track of the user that created a model instance.
    This will only be set once upon an INSERT in the database.
    """

    # dont actually need to do anything, everything is handled by the parent class
    # the different logic goes in the middleware
    pass


class CreatingSessionKeyField(LastSessionKeyField):
    """
    A field that keeps track of the last session key with which a model instance was created.
    This will only be set once upon an INSERT in the database.
    """

    # dont actually need to do anything, everything is handled by the parent class
    # the different logic goes in the middleware
    pass


# South stuff:

rules = [
    (
        (LastUserField, CreatingUserField),
        [],
        {
            "to": [
                "rel.to",
                {"default": getattr(settings, "AUTH_USER_MODEL", "auth.User")},
            ],
            "null": ["null", {"default": True}],
        },
    )
]

try:
    from south.modelsinspector import add_introspection_rules

    # Add the rules for the `LastUserField`
    add_introspection_rules(rules, ["^audit_log\.models\.fields\.LastUserField"])
    add_introspection_rules(rules, ["^audit_log\.models\.fields\.CreatingUserField"])
    add_introspection_rules([], ["^audit_log\.models\.fields\.LastSessionKeyField"])
    add_introspection_rules([], ["^audit_log\.models\.fields\.CreatingSessionKeyField"])
except ImportError:
    pass
